@@ -47,8 +47,6 @@ let debug_backtraces =
 
 let default_build_dir = "_build"
 
-(* Allow options from term1 or exclusively options from term2. If the user
-   passes options from both terms, an error is reported. *)
 let one_of term1 term2 =
   Term.ret
   @@ let+ x, args1 = Term.with_used_args term1
@@ -102,7 +100,7 @@ module Options_implied_by_dash_p = struct
     ; always_show_command_line : bool
     ; promote_install_files : bool
     ; require_dune_project_file : bool
-    ; ignore_lock_directory : bool
+    ; ignore_lock_dir : bool
     }
 
   let docs = copts_sect
@@ -216,9 +214,9 @@ module Options_implied_by_dash_p = struct
         last
         & opt_all ~vopt:true bool [ false ]
         & info [ "require-dune-project-file" ] ~docs ~doc)
-    and+ ignore_lock_directory =
+    and+ ignore_lock_dir =
       let doc = "Ignore dune.lock/ directory." in
-      Arg.(value & flag & info [ "ignore-lock-directory" ] ~docs ~doc)
+      Arg.(value & flag & info [ "ignore-lock-dir" ] ~docs ~doc)
     in
     { root
     ; only_packages = No_restriction
@@ -229,7 +227,7 @@ module Options_implied_by_dash_p = struct
     ; always_show_command_line
     ; promote_install_files
     ; require_dune_project_file
-    ; ignore_lock_directory
+    ; ignore_lock_dir
     }
   ;;
 
@@ -244,7 +242,7 @@ module Options_implied_by_dash_p = struct
       ; "--always-show-command-line"
       ; "--promote-install-files"
       ; "--require-dune-project-file"
-      ; "--ignore-lock-directory"
+      ; "--ignore-lock-dir"
       ; "--default-target"
       ; "@install"
       ]
@@ -577,7 +575,7 @@ module Builder = struct
     ; ignore_promoted_rules : bool
     ; force : bool
     ; no_print_directory : bool
-    ; ignore_lock_directory : bool
+    ; ignore_lock_dir : bool
     ; store_orig_src_dir : bool
     ; default_target : Arg.Dep.t (* For build & runtest only *)
     ; watch : Dune_rpc_impl.Watch_mode_config.t
@@ -595,10 +593,8 @@ module Builder = struct
     ; separate_error_messages : bool
     ; stop_on_first_error : bool
     ; require_dune_project_file : bool
-    ; insignificant_changes : [ `React | `Ignore ]
     ; watch_exclusions : string list
     ; build_dir : string
-    ; store_digest_preimage : bool
     ; root : string option
     ; stats_trace_file : string option
     ; stats_trace_extended : bool
@@ -670,19 +666,6 @@ module Builder = struct
             [ "debug-digests" ]
             ~docs
             ~doc:"Explain why Dune decides to re-digest some files")
-    and+ store_digest_preimage =
-      Arg.(
-        value
-        & flag
-        & info
-            [ "debug-store-digest-preimage" ]
-            ~docs
-            ~doc:
-              "Store digest preimage for all computed digests, so that it's possible to \
-               reverse them later, for debugging. The digests are stored in the shared \
-               cache (see --cache flag) as values, even if cache is otherwise disabled. \
-               This should be used only for debugging, since it's slow and it litters \
-               the shared cache.")
     and+ no_buffer =
       let doc =
         "Do not buffer the output of commands executed by dune. By default dune buffers \
@@ -820,7 +803,7 @@ module Builder = struct
          ; always_show_command_line
          ; promote_install_files
          ; require_dune_project_file
-         ; ignore_lock_directory
+         ; ignore_lock_dir
          }
       =
       Options_implied_by_dash_p.term
@@ -965,15 +948,6 @@ module Builder = struct
                of the build in a deterministic order. $(b,twice) reports each error \
                twice: once as soon as the error is discovered and then again at the end \
                of the build, in a deterministic order.")
-    and+ react_to_insignificant_changes =
-      Arg.(
-        value
-        & flag
-        & info
-            [ "react-to-insignificant-changes" ]
-            ~doc:
-              "React to insignificant file system changes; this is only useful for \
-               benchmarking dune.")
     and+ separate_error_messages =
       Arg.(
         value
@@ -1004,7 +978,7 @@ module Builder = struct
     ; ignore_promoted_rules
     ; force
     ; no_print_directory
-    ; ignore_lock_directory
+    ; ignore_lock_dir
     ; store_orig_src_dir
     ; default_target
     ; watch
@@ -1035,10 +1009,8 @@ module Builder = struct
     ; separate_error_messages
     ; stop_on_first_error
     ; require_dune_project_file
-    ; insignificant_changes = (if react_to_insignificant_changes then `React else `Ignore)
     ; watch_exclusions
     ; build_dir = Option.value ~default:default_build_dir build_dir
-    ; store_digest_preimage
     ; root
     ; stats_trace_file
     ; stats_trace_extended
@@ -1075,17 +1047,8 @@ let rpc t =
   | `Allow rpc -> `Allow (Lazy.force rpc)
 ;;
 
-let signal_watcher t =
-  match t.rpc with
-  | `Allow _ -> `Yes
-  | `Forbid_builds ->
-    (* if we aren't building anything, then we don't mind interrupting dune immediately *)
-    `No
-;;
-
 let watch_exclusions t = t.builder.watch_exclusions
 let stats t = t.stats
-let insignificant_changes t = t.builder.insignificant_changes
 
 (* To avoid needless recompilations under Windows, where the case of
    [Sys.getcwd] can vary between different invocations of [dune], normalize to
@@ -1196,7 +1159,6 @@ let build (builder : Builder.t) =
              action_runner))
     else `Forbid_builds
   in
-  if builder.store_digest_preimage then Dune_engine.Reversible_digest.enable ();
   if builder.print_metrics then Dune_metrics.enable ();
   { builder; root; rpc; stats }
 ;;
@@ -1272,9 +1234,9 @@ let init (builder : Builder.t) =
   Dune_engine.Clflags.debug_backtraces c.builder.debug_backtraces;
   Dune_rules.Clflags.debug_artifact_substitution := c.builder.debug_artifact_substitution;
   Dune_engine.Clflags.debug_load_dir := c.builder.debug_load_dir;
-  Dune_engine.Clflags.debug_digests := c.builder.debug_digests;
   Dune_engine.Clflags.debug_fs_cache := c.builder.cache_debug_flags.fs_cache;
-  Dune_engine.Clflags.wait_for_filesystem_clock := c.builder.wait_for_filesystem_clock;
+  Dune_digest.Clflags.debug_digests := c.builder.debug_digests;
+  Dune_digest.Clflags.wait_for_filesystem_clock := c.builder.wait_for_filesystem_clock;
   Dune_engine.Clflags.capture_outputs := c.builder.capture_outputs;
   Dune_engine.Clflags.diff_command := c.builder.diff_command;
   Dune_engine.Clflags.promote := c.builder.promote;
@@ -1284,7 +1246,7 @@ let init (builder : Builder.t) =
   Dune_rules.Clflags.promote_install_files := c.builder.promote_install_files;
   Dune_engine.Clflags.always_show_command_line := c.builder.always_show_command_line;
   Dune_rules.Clflags.ignore_promoted_rules := c.builder.ignore_promoted_rules;
-  Dune_rules.Clflags.ignore_lock_directory := c.builder.ignore_lock_directory;
+  Dune_rules.Clflags.ignore_lock_dir := c.builder.ignore_lock_dir;
   Dune_rules.Clflags.on_missing_dune_project_file
   := if c.builder.require_dune_project_file then Error else Warn;
   Log.info
