@@ -274,12 +274,13 @@ module Sanitize_for_tests = struct
     let fake_workspace = lazy (Path.External.of_string "/WORKSPACE_ROOT")
 
     let sanitize_with_findlib ~findlib_paths path =
+      let path = Path.external_ path in
       List.find_map findlib_paths ~f:(fun candidate ->
         let open Option.O in
         let* candidate = Path.as_external candidate in
         (* if the path to rename is an external path, try to find the
            OCaml root inside, and replace it with a fixed string *)
-        let+ without_prefix = Path.External.drop_prefix ~prefix:candidate path in
+        let+ without_prefix = Path.drop_prefix ~prefix:(Path.external_ candidate) path in
         (* we have found the OCaml root path: let's replace it with a
            constant string *)
         Path.External.append_local (Lazy.force fake_findlib) without_prefix)
@@ -371,7 +372,9 @@ module Crawl = struct
 
   (* Builds the list of modules *)
   let modules ~obj_dir ~deps_of modules_ : Descr.Mod.t list Memo.t =
-    Modules.fold_no_vlib ~init:(Memo.return []) modules_ ~f:(fun m macc ->
+    modules_
+    |> Modules.With_vlib.drop_vlib
+    |> Modules.fold ~init:(Memo.return []) ~f:(fun m macc ->
       let* acc = macc in
       let deps = deps_of m in
       let+ { Ocaml.Ml_kind.Dict.intf = deps_for_intf; impl = deps_for_impl }, _ =
@@ -384,11 +387,19 @@ module Crawl = struct
   let executables sctx ~options ~project ~dir (exes : Executables.t)
     : (Descr.Item.t * Lib.Set.t) option Memo.t
     =
-    let first_exe = snd (List.hd exes.names) in
+    let first_exe = snd (Nonempty_list.hd exes.names) in
+    let* scope =
+      Scope.DB.find_by_project (Super_context.context sctx |> Context.name) project
+    in
     let* modules_, obj_dir =
-      Dir_contents.get sctx ~dir
-      >>= Dir_contents.ocaml
-      >>| Ml_sources.modules_and_obj_dir ~for_:(Exe { first_exe })
+      let+ modules_, obj_dir =
+        Dir_contents.get sctx ~dir
+        >>= Dir_contents.ocaml
+        >>= Ml_sources.modules_and_obj_dir
+              ~libs:(Scope.libs scope)
+              ~for_:(Exe { first_exe })
+      in
+      Modules.With_vlib.modules modules_, obj_dir
     in
     let* pp_map =
       let+ version =
@@ -396,7 +407,7 @@ module Crawl = struct
         ocaml.version
       in
       Staged.unstage
-      @@ Preprocessing.pped_modules_map
+      @@ Pp_spec.pped_modules_map
            (Preprocess.Per_module.without_instrumentation exes.buildable.preprocess)
            version
     in
@@ -405,9 +416,6 @@ module Crawl = struct
       immediate_deps_of_module ~options ~obj_dir ~modules:modules_ module_
     in
     let obj_dir = Obj_dir.of_local obj_dir in
-    let* scope =
-      Scope.DB.find_by_project (Super_context.context sctx |> Context.name) project
-    in
     let* modules_ = modules ~obj_dir ~deps_of modules_ in
     let+ requires =
       let* compile_info = Exe_rules.compile_info ~scope exes in
@@ -424,7 +432,7 @@ module Crawl = struct
     | Ok libs ->
       let include_dirs = Obj_dir.all_cmis obj_dir in
       let exe_descr =
-        { Descr.Exe.names = List.map ~f:snd exes.names
+        { Descr.Exe.names = List.map ~f:snd (Nonempty_list.to_list exes.names)
         ; requires = List.map ~f:uid_of_library libs
         ; modules = modules_
         ; include_dirs
@@ -449,9 +457,17 @@ module Crawl = struct
         | true ->
           (* XXX why do we have a second object directory? *)
           let* modules_, obj_dir_ =
-            Dir_contents.get sctx ~dir:(Path.as_in_build_dir_exn src_dir)
-            >>= Dir_contents.ocaml
-            >>| Ml_sources.modules_and_obj_dir ~for_:(Library name)
+            let* libs =
+              Scope.DB.find_by_dir (Path.as_in_build_dir_exn src_dir) >>| Scope.libs
+            in
+            let+ modules_, obj_dir_ =
+              Dir_contents.get sctx ~dir:(Path.as_in_build_dir_exn src_dir)
+              >>= Dir_contents.ocaml
+              >>= Ml_sources.modules_and_obj_dir
+                    ~libs
+                    ~for_:(Library (Lib_info.lib_id info |> Lib_id.to_local_exn))
+            in
+            Modules.With_vlib.modules modules_, obj_dir_
           in
           let* pp_map =
             let+ version =
@@ -459,7 +475,7 @@ module Crawl = struct
               ocaml.version
             in
             Staged.unstage
-            @@ Preprocessing.pped_modules_map
+            @@ Pp_spec.pped_modules_map
                  (Preprocess.Per_module.without_instrumentation
                     (Lib_info.preprocess info))
                  version
