@@ -84,7 +84,6 @@ let local_libraries =
   ; ("vendor/lwd/nottui", None, false, None)
   ; ("src/dune_tui", Some "Dune_tui", true, None)
   ; ("src/dune_config_file", Some "Dune_config_file", false, None)
-  ; ("src/dune_shared_cache", Some "Dune_shared_cache", false, None)
   ; ("src/dune_patch", Some "Dune_patch", false, None)
   ; ("src/scheme", Some "Scheme", false, None)
   ; ("src/dune_rules", Some "Dune_rules", true, None)
@@ -96,26 +95,18 @@ let local_libraries =
   ]
 
 let build_flags =
-  [ ("win32", [ "-ccopt"; "-D_UNICODE"; "-ccopt"; "-DUNICODE" ])
-  ; ("win64", [ "-ccopt"; "-D_UNICODE"; "-ccopt"; "-DUNICODE" ])
-  ; ("mingw", [ "-ccopt"; "-D_UNICODE"; "-ccopt"; "-DUNICODE" ])
-  ; ("mingw64", [ "-ccopt"; "-D_UNICODE"; "-ccopt"; "-DUNICODE" ])
+  [ ([ "win32"; "win64"; "mingw"; "mingw64" ],
+    [ "-ccopt"; "-D_UNICODE"; "-ccopt"; "-DUNICODE" ])
   ]
 
 let link_flags =
-  [ ("macosx",
+  [ ([ "macosx" ],
     [ "-cclib"
     ; "-framework CoreFoundation"
     ; "-cclib"
     ; "-framework CoreServices"
     ])
-  ; ("win32",
+  ; ([ "win32"; "win64"; "mingw"; "mingw64" ],
     [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ])
-  ; ("win64",
-    [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ])
-  ; ("mingw",
-    [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ])
-  ; ("mingw64",
-    [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ])
-  ; ("beos", [ "-cclib"; "-lbsd" ])
+  ; ([ "beos" ], [ "-cclib"; "-lbsd" ])
   ]
