@@ -279,6 +279,111 @@ let%expect_test _ =
 |}]
 ;;
 
+let%expect_test _ =
+  reach "foo" ~from:"foo";
+  [%expect {|
+"."
+|}]
+;;
+
+let%expect_test _ =
+  reach "bar/foo" ~from:"bar/foo";
+  [%expect {|
+"."
+|}]
+;;
+
+let%expect_test _ =
+  reach "a/b/x" ~from:"a/b/y";
+  [%expect {|
+"../x"
+|}]
+;;
+
+let%expect_test _ =
+  reach "a/b" ~from:"a/b/x";
+  [%expect {|
+".."
+|}]
+;;
+
+let%expect_test _ =
+  reach "a/b/x" ~from:"a/b";
+  [%expect {|
+"x"
+|}]
+;;
+
+let%expect_test _ =
+  reach "a/b/x/z" ~from:"a/b/y";
+  [%expect {|
+"../x/z"
+|}]
+;;
+
+let%expect_test _ =
+  reach "a/b/y" ~from:"a/b/x/z";
+  [%expect {|
+"../../y"
+|}]
+;;
+
+let%expect_test _ =
+  reach "a/bbb" ~from:"a/b";
+  [%expect {|
+"../bbb"
+|}]
+;;
+
+let%expect_test _ =
+  reach "" ~from:"";
+  [%expect {|
+"."
+|}]
+;;
+
+let%expect_test _ =
+  reach "" ~from:"foo";
+  [%expect {|
+".."
+|}]
+;;
+
+let%expect_test _ =
+  reach "foo" ~from:"";
+  [%expect {|
+"foo"
+|}]
+;;
+
+let%expect_test _ =
+  reach "x/foo" ~from:"bar/x";
+  [%expect {|
+"../../x/foo"
+|}]
+;;
+
+let%expect_test _ =
+  reach "a/x" ~from:"x/b";
+  [%expect {|
+"../../a/x"
+|}]
+;;
+
+let%expect_test _ =
+  reach "default/META.foo" ~from:"default";
+  [%expect {|
+"META.foo"
+|}]
+;;
+
+let%expect_test _ =
+  reach "default/av" ~from:"default/avdevice";
+  [%expect {|
+"../av"
+|}]
+;;
+
 let%expect_test _ =
   relative (Path.of_string "relative") "/absolute/path";
   [%expect {|
@@ -469,9 +574,9 @@ let%expect_test "drop prefix" =
 ;;
 
 let%expect_test "drop external prefix" =
-  Path.External.drop_prefix
-    ~prefix:(Path.External.of_filename_relative_to_initial_cwd "foo/bar")
-    (Path.External.of_filename_relative_to_initial_cwd "foo/bar/baz")
+  Path.drop_prefix
+    ~prefix:(Path.of_filename_relative_to_initial_cwd "foo/bar")
+    (Path.of_filename_relative_to_initial_cwd "foo/bar/baz")
   |> Dyn.option Path.Local.to_dyn
   |> print_dyn;
   [%expect {| Some "baz" |}]
@@ -485,9 +590,9 @@ let%expect_test "drop prefix as substring" =
 ;;
 
 let%expect_test "drop external prefix as substring" =
-  Path.External.drop_prefix
-    ~prefix:(Path.External.of_filename_relative_to_initial_cwd "foo/bar")
-    (Path.External.of_filename_relative_to_initial_cwd "foo/barbaz")
+  Path.drop_prefix
+    ~prefix:(Path.of_filename_relative_to_initial_cwd "foo/bar")
+    (Path.of_filename_relative_to_initial_cwd "foo/barbaz")
   |> Dyn.option Path.Local.to_dyn
   |> print_dyn;
   [%expect {| None |}]
@@ -500,7 +605,14 @@ let%expect_test "drop entire path" =
 ;;
 
 let%expect_test "drop entire external path" =
-  let path = Path.External.of_filename_relative_to_initial_cwd "foo/bar" in
-  Path.External.drop_prefix ~prefix:path path |> Dyn.option Path.Local.to_dyn |> print_dyn;
+  let path = Path.of_filename_relative_to_initial_cwd "foo/bar" in
+  Path.drop_prefix ~prefix:path path |> Dyn.option Path.Local.to_dyn |> print_dyn;
   [%expect {| Some "." |}]
 ;;
+
+let%expect_test "drop prefix with a trailing /" =
+  Path.drop_prefix ~prefix:(Path.of_string "/a/b/c/") (Path.of_string "/a/b/c/d/e")
+  |> Dyn.option Path.Local.to_dyn
+  |> print_dyn;
+  [%expect {| Some "d/e" |}]
+;;
