@@ -3,31 +3,11 @@
 open! Import
 
 module Config : sig
-  module Display : sig
-    type verbosity =
-      | Quiet  (** Only display errors *)
-      | Short  (** One line per command *)
-      | Verbose  (** Display all commands fully *)
-
-    type t =
-      { status_line : bool
-      ; verbosity : verbosity
-      }
-
-    val all : (string * t) list
-
-    val to_dyn : t -> Dyn.t
-
-    (** The console backend corresponding to the selected display mode *)
-    val console_backend : t -> Console.Backend.t
-  end
-
   type t =
     { concurrency : int
-    ; display : Display.t
     ; stats : Dune_stats.t option
-    ; insignificant_changes : [ `Ignore | `React ]
-    ; signal_watcher : [ `Yes | `No ]
+    ; print_ctrl_c_warning : bool
+    ; watch_exclusions : string list
     }
 end
 
@@ -51,17 +31,10 @@ module Run : sig
     | No_watcher
 
   module Shutdown : sig
-    module Signal : sig
-      (* TODO move this stuff into stdune? *)
-      type t =
-        | Int
-        | Quit
-        | Term
-    end
-
     module Reason : sig
       type t =
         | Requested
+        | Timeout
         | Signal of Signal.t
     end
 
@@ -88,12 +61,12 @@ module Run : sig
 
       The fiber [get_build_request] is run at the beginning of every iteration
       to wait for the build signal. *)
-  val poll_passive :
-       get_build_request:(step * Build_outcome.t Fiber.Ivar.t) Fiber.t
+  val poll_passive
+    :  get_build_request:(step * Build_outcome.t Fiber.Ivar.t) Fiber.t
     -> unit Fiber.t
 
-  val go :
-       Config.t
+  val go
+    :  Config.t
     -> ?timeout:float
     -> ?file_watcher:file_watcher
     -> on_event:(Config.t -> Event.t -> unit)
@@ -101,23 +74,11 @@ module Run : sig
     -> 'a
 end
 
-module Worker : sig
-  (** A worker is a thread that runs submitted tasks *)
-  type t
+(** [async f] runs [f] inside a background thread pool *)
+val async : (unit -> 'a) -> ('a, Exn_with_backtrace.t) result Fiber.t
 
-  val create : unit -> t Fiber.t
-
-  val task :
-       t
-    -> f:(unit -> 'a)
-    -> ('a, [ `Exn of Exn_with_backtrace.t | `Stopped ]) result Fiber.t
-
-  (** Should be used for tasks never raise and always complete before stop is
-      called *)
-  val task_exn : t -> f:(unit -> 'a) -> 'a Fiber.t
-
-  val stop : t -> unit
-end
+(** [async_exn f] runs [f] inside a background thread pool *)
+val async_exn : (unit -> 'a) -> 'a Fiber.t
 
 type t
 
@@ -132,17 +93,21 @@ val with_job_slot : (Fiber.Cancel.t -> Config.t -> 'a Fiber.t) -> 'a Fiber.t
 (** Wait for the following process to terminate. If [is_process_group_leader] is
     true, kill the entire process group instead of just the process in case of
     timeout. *)
-val wait_for_process :
-     ?timeout:float
+val wait_for_process
+  :  ?timeout:float
   -> ?is_process_group_leader:bool
   -> Pid.t
   -> Proc.Process_info.t Fiber.t
 
-val yield_if_there_are_pending_events : unit -> unit Fiber.t
+type termination_reason =
+  | Normal
+  | Cancel
 
-(** If the current build was cancelled, raise
-    [Memo.Non_reproducible Run.Build_cancelled]. *)
-val abort_if_build_was_cancelled : unit Fiber.t
+val wait_for_build_process
+  :  ?timeout:float
+  -> ?is_process_group_leader:bool
+  -> Pid.t
+  -> (Proc.Process_info.t * termination_reason) Fiber.t
 
 (** Number of jobs currently running in the background *)
 val running_jobs_count : t -> int
@@ -170,6 +135,12 @@ val running_jobs_count : t -> int
     will get suspended and will never restart. *)
 val shutdown : unit -> unit Fiber.t
 
+(** Cancel the current build. Superficially, this function is like [shutdown]
+    in that it stops the build early, but it is different because the [Run.go]
+    call is allowed to complete its fiber. In this respect, the behavior is
+    similar to what happens on file system events in polling mode. *)
+val cancel_current_build : unit -> unit Fiber.t
+
 val inject_memo_invalidation : Memo.Invalidation.t -> unit Fiber.t
 
 (** [sleep duration] wait for [duration] to elapse. Sleepers are checked for
@@ -186,3 +157,5 @@ val stats : unit -> Dune_stats.t option Fiber.t
     clients may observe that Dune reacted to a file change. This is needed for
     benchmarking the watch mode of Dune. *)
 val wait_for_build_input_change : unit -> unit Fiber.t
+
+val spawn_thread : (unit -> unit) -> unit
