@@ -16,6 +16,7 @@ module Package_constraint = Package_constraint
 module Package_name = Package_name
 module Package_dependency = Package_dependency
 module Package_version = Package_version
+module Package_variable_name = Package_variable_name
 module Lib_kind = Lib_kind
 module Lib_dep = Lib_dep
 module Pkg = Pkg
@@ -30,3 +31,12 @@ module Targets_spec = Targets_spec
 module Wrapped = Wrapped
 module Visibility = Visibility
 module Dep_conf = Dep_conf
+module Relop = Relop
+module Cond = Cond
+module Toggle = Toggle
+module Site = Site
+module Warning = Warning
+module Source_kind = Source_kind
+module Package_info = Package_info
+module Section = Section
+module Package = Package
