@@ -2,9 +2,10 @@ open! Stdune
 
 module Constraint : sig
   module Op : sig
-    type t = Dune_lang.Package_constraint.Op.t
+    type t = Dune_lang.Relop.t
 
-    val to_relop : t -> OpamParserTypes.FullPos.relop
+    val to_opam : t -> OpamParserTypes.relop
+    val of_opam : OpamParserTypes.relop -> t
   end
 
   module Value : sig
@@ -25,3 +26,7 @@ include module type of Dune_lang.Package_dependency with type t := t
 
 val opam_depend : t -> OpamParserTypes.FullPos.value
 val list_to_opam_filtered_formula : t list -> OpamTypes.filtered_formula
+
+(** Attempt to interpret a [OpamTypes.filtered_formula] as a list of [t]s by
+    treating the formula as a conjunction of packages with constraints. *)
+val list_of_opam_filtered_formula : Loc.t -> OpamTypes.filtered_formula -> t list
