@@ -85,20 +85,19 @@ let with_flock lock_path ~f =
       [ Unix.O_CREAT; O_WRONLY; O_SHARE_DELETE; Unix.O_CLOEXEC ]
       0o600
   in
-  let out = Unix.out_channel_of_descr fd in
   let flock = Flock.create fd in
   let max_retries = 49 in
   Fiber.finalize
     ~finally:(fun () ->
       let+ () = Fiber.return () in
-      close_out out)
+      Unix.close fd)
     (fun () ->
       attempt_to_lock flock Flock.Exclusive ~max_retries
       >>= function
       | Ok `Success ->
         Fiber.finalize
           (fun () ->
-            Printf.fprintf out "%d\n%!" (Unix.getpid ());
+            Dune_util.Global_lock.write_pid fd;
             f ())
           ~finally:(fun () ->
             let+ () = Fiber.return () in
@@ -134,21 +133,39 @@ let output_limit = Sys.max_string_length
 let make_stdout () = Process.Io.make_stdout ~output_on_success:Swallow ~output_limit
 let make_stderr () = Process.Io.make_stderr ~output_on_success:Swallow ~output_limit
 
-(* to avoid Git translating its CLI *)
-let env = Env.add Env.initial ~var:"LC_ALL" ~value:"C"
-
-let git_code_error ~dir ~args ~exit_code ~output =
-  let git = Lazy.force Vcs.git in
-  Code_error.raise
-    "git returned non-zero exit code"
-    [ "exit code", Dyn.int exit_code
-    ; "dir", Path.to_dyn dir
-    ; "git", Path.to_dyn git
-    ; "args", Dyn.list Dyn.string args
-    ; "output", Dyn.list Dyn.string output
-    ]
+let env =
+  (* to avoid Git translating its CLI *)
+  Env.add Env.initial ~var:"LC_ALL" ~value:"C"
+  (* to avoid prmompting for passwords *)
+  |> Env.add ~var:"GIT_TERMINAL_PROMPT" ~value:"0"
 ;;
 
+module Git_error = struct
+  type t =
+    { dir : Path.t
+    ; args : string list
+    ; exit_code : int
+    ; output : string list
+    }
+
+  let raise_code_error { dir; args; exit_code; output } =
+    let git = Lazy.force Vcs.git in
+    Code_error.raise
+      "git returned non-zero exit code"
+      [ "exit code", Dyn.int exit_code
+      ; "dir", Path.to_dyn dir
+      ; "git", Path.to_dyn git
+      ; "args", Dyn.list Dyn.string args
+      ; "output", Dyn.list Dyn.string output
+      ]
+  ;;
+
+  let result_get_or_code_error = function
+    | Ok x -> x
+    | Error t -> raise_code_error t
+  ;;
+end
+
 let run_with_exit_code { dir; _ } ~allow_codes ~display args =
   let stdout_to = make_stdout () in
   let stderr_to = make_stderr () in
@@ -157,12 +174,13 @@ let run_with_exit_code { dir; _ } ~allow_codes ~display args =
     Process.run ~dir ~display ~stdout_to ~stderr_to ~env failure_mode git args
   in
   if allow_codes exit_code
-  then exit_code
-  else git_code_error ~dir ~args ~exit_code ~output:[]
+  then Ok exit_code
+  else Error { Git_error.dir; args; exit_code; output = [] }
 ;;
 
 let run t ~display args =
-  run_with_exit_code t ~allow_codes:(Int.equal 0) ~display args >>| ignore
+  run_with_exit_code t ~allow_codes:(Int.equal 0) ~display args
+  >>| Result.map ~f:(ignore : int -> unit)
 ;;
 
 let run_capture_lines { dir; _ } ~display args =
@@ -170,7 +188,7 @@ let run_capture_lines { dir; _ } ~display args =
   let+ output, exit_code =
     Process.run_capture_lines ~dir ~display ~env failure_mode git args
   in
-  if exit_code = 0 then output else git_code_error ~dir ~args ~exit_code ~output
+  if exit_code = 0 then Ok output else Error { Git_error.dir; args; exit_code; output }
 ;;
 
 let run_capture_zero_separated_lines { dir; _ } args =
@@ -178,7 +196,7 @@ let run_capture_zero_separated_lines { dir; _ } args =
   let+ output, exit_code =
     Process.run_capture_zero_separated ~dir ~display:Quiet ~env failure_mode git args
   in
-  if exit_code = 0 then output else git_code_error ~dir ~args ~exit_code ~output
+  if exit_code = 0 then Ok output else Error { Git_error.dir; args; exit_code; output }
 ;;
 
 let cat_file { dir; _ } command =
@@ -297,7 +315,11 @@ let load_or_create ~dir =
     with_flock lock ~f:(fun () ->
       match Fpath.mkdir_p (Path.to_string dir) with
       | Already_exists -> Fiber.return ()
-      | Created -> run t ~display:Quiet [ "init"; "--bare" ]
+      | Created ->
+        run t ~display:Quiet [ "init"; "--bare" ]
+        >>| (function
+         | Ok () -> ()
+         | Error git_error -> Git_error.raise_code_error git_error)
       | exception Unix.Unix_error (e, x, y) ->
         User_error.raise
           [ Pp.textf "%s isn't a directory" (Path.to_string_maybe_quoted dir)
@@ -457,10 +479,11 @@ let fetch_allow_failure repo ~url obj =
         ~display:!Dune_engine.Clflags.display
         [ "fetch"; "--no-write-fetch-head"; url; Object.to_string obj ]
       >>| (function
-       | 128 -> `Not_found
-       | 0 ->
+       | Ok 128 -> `Not_found
+       | Ok 0 ->
          Table.set repo.present_objects obj ();
          `Fetched
+       | Error git_error -> Git_error.raise_code_error git_error
        | _ -> assert false))
 ;;
 
@@ -521,6 +544,7 @@ module At_rev = struct
     let config repo (Object.Sha1 rev) path : t Fiber.t =
       [ "config"; "--list"; "--blob"; sprintf "%s:%s" rev (Path.Local.to_string path) ]
       |> run_capture_lines repo ~display:Quiet
+      >>| Git_error.result_get_or_code_error
       >>| List.fold_left ~init:KV.Map.empty ~f:(fun acc line ->
         match parse line with
         | None ->
@@ -577,6 +601,7 @@ module At_rev = struct
 
   let files_and_submodules repo (Object.Sha1 rev) =
     run_capture_zero_separated_lines repo [ "ls-tree"; "-z"; "--long"; "-r"; rev ]
+    >>| Git_error.result_get_or_code_error
     >>| List.fold_left
           ~init:(File.Set.empty, Commit.Set.empty)
           ~f:(fun (files, commits) line ->
@@ -676,7 +701,7 @@ module At_rev = struct
   let check_out { repo = { dir; _ }; revision = Sha1 rev; files = _ } ~target =
     (* TODO iterate over submodules to output sources *)
     let git = Lazy.force Vcs.git in
-    let temp_dir = Temp.create Dir ~prefix:"rev-store" ~suffix:rev in
+    let temp_dir = Temp_dir.dir_for_target ~target ~prefix:"rev-store" ~suffix:rev in
     Fiber.finalize ~finally:(fun () ->
       let+ () = Fiber.return () in
       Temp.destroy Dir temp_dir)
@@ -689,7 +714,8 @@ module At_rev = struct
       let+ (), exit_code =
         Process.run ~dir ~display:Quiet ~stdout_to ~stderr_to ~env failure_mode git args
       in
-      if exit_code <> 0 then git_code_error ~dir ~args ~exit_code ~output:[]
+      if exit_code <> 0
+      then Git_error.raise_code_error { dir; args; exit_code; output = [] }
     in
     (* We untar things into a temp dir to make sure we don't create garbage
        in the build dir until we know can produce the files *)
@@ -724,12 +750,31 @@ let remote =
                ]
            ])
   in
-  fun t ~url ->
+  fun t ~url:(url_loc, url) ->
     let f url =
       let command = [ "ls-remote"; url ] in
       let refs =
         Fiber_lazy.create (fun () ->
-          let+ hits = run_capture_lines t ~display:!Dune_engine.Clflags.display command in
+          let+ hits =
+            run_capture_lines t ~display:!Dune_engine.Clflags.display command
+            >>| function
+            | Ok lines -> lines
+            | Error git_error ->
+              (match git_error.exit_code with
+               | 128 ->
+                 User_error.raise
+                   ~loc:url_loc
+                   ~hints:
+                     [ Pp.textf
+                         "Check that this Git URL in the project configuration is \
+                          correct: %S"
+                         url
+                     ]
+                   [ Pp.text "Failed to run external command:"
+                   ; User_message.command (sprintf "git ls-remote %S" url)
+                   ]
+               | _ -> Git_error.raise_code_error git_error)
+          in
           let default_branch, branches_and_tags =
             List.fold_left
               hits
