@@ -3,11 +3,12 @@ open Import
 module Repository : sig
   type t
 
-  val opam_url : t -> OpamUrl.t
+  val opam_url : t -> Loc.t * OpamUrl.t
   val hash : t -> int
   val to_dyn : t -> Dyn.t
   val equal : t -> t -> bool
-  val default : t
+  val upstream : t
+  val overlay : t
   val decode : t Decoder.t
 
   module Name : sig
@@ -21,5 +22,4 @@ module Repository : sig
   end
 
   val name : t -> Name.t
-  val create : name:Name.t -> source:string -> t
 end
