@@ -5,6 +5,11 @@ open Import
 let ( let+! ) = Memo.O.( let+ )
 let ( let*! ) = Memo.O.( let* )
 
+let pp_targets targets =
+  Pp.enumerate (Targets.all targets) ~f:(fun target ->
+    Pp.text (Dpath.describe_target target))
+;;
+
 let validate_target_dir ~targets_dir ~loc targets path =
   if Path.Build.(parent_exn path <> targets_dir)
   then
@@ -13,7 +18,7 @@ let validate_target_dir ~targets_dir ~loc targets path =
       [ Pp.text
           "This action has targets in a different directory than the current one, this \
            is not allowed by dune at the moment:"
-      ; Targets.pp targets
+      ; pp_targets targets
       ]
 ;;
 
@@ -45,7 +50,7 @@ module Action_expander : sig
     -> expander:Expander.t
     -> 'a Action_builder.With_targets.t Memo.t
 
-  val with_expander : (Expander.t -> 'a t) -> 'a t
+  val with_expander : (Expander.t -> 'a t Memo.t) -> 'a t
 
   (* String with vars expansion *)
   module E : sig
@@ -105,8 +110,9 @@ end = struct
 
   let return x _env acc = Memo.return (Action_builder.return x, acc)
 
-  let with_expander (type a) (f : Expander.t -> a t) env acc =
-    let f = f env.expander in
+  let with_expander (type a) (f : Expander.t -> a t Memo.t) env acc =
+    let open Memo.O in
+    let* f = f env.expander in
     f env acc
   ;;
 
@@ -391,7 +397,15 @@ end = struct
                Action_builder.return (Ok (Path.relative dir s))
              | In_path ->
                Action_builder.of_memo
-                 (Artifacts.binary ~loc:(Some loc) (Expander.artifacts env.expander) s))
+               @@
+               let open Memo.O in
+               let* where =
+                 let+ project = Dune_load.find_project ~dir:env.dir in
+                 if Dune_project.dune_version project >= (3, 14)
+                 then Artifacts.Original_path
+                 else Install_dir
+               and* artifacts = Expander.artifacts env.expander in
+               Artifacts.binary ~loc:(Some loc) ~where artifacts s)
         in
         let prog = Result.map prog ~f:(Expander.map_exe env.expander) in
         let args = Value.L.to_strings ~dir args in
@@ -471,9 +485,12 @@ let rec expand (t : Dune_lang.Action.t) : Action.t Action_expander.t =
     O.Echo l
   | Cat xs ->
     A.with_expander (fun expander ->
-      let version =
-        Expander.scope expander |> Scope.project |> Dune_project.dune_version
+      let open Memo.O in
+      let+ version =
+        let dir = Expander.dir expander in
+        Dune_load.find_project ~dir >>| Dune_project.dune_version
       in
+      let open Action_expander.O in
       if version >= (3, 10)
       then
         let+ xs = A.all (List.map xs ~f:E.deps) in
@@ -491,10 +508,12 @@ let rec expand (t : Dune_lang.Action.t) : Action.t Action_expander.t =
     O.Symlink (x, y)
   | Copy_and_add_line_directive (x, y) ->
     A.with_expander (fun expander ->
-      let context = Expander.context expander in
-      let+ x = E.dep x
-      and+ y = E.target y in
-      Copy_line_directive.action context ~src:x ~dst:y)
+      Expander.context expander
+      |> Context.DB.get
+      |> Memo.map ~f:(fun context ->
+        let+ x = E.dep x
+        and+ y = E.target y in
+        Copy_line_directive.action context ~src:x ~dst:y))
   | System x ->
     let+ x = E.string x in
     O.System x
@@ -567,7 +586,7 @@ let expand_no_targets t ~loc ~chdir ~deps:deps_written_by_user ~expander ~what =
           "%s must not have targets, however I inferred that these files will be created \
            by this action:"
           (String.capitalize what)
-      ; Targets.pp targets
+      ; pp_targets targets
       ];
   let+ () = deps_builder
   and+ action = build in
