@@ -5,7 +5,7 @@ let def name dyn =
   Pp.box ~indent:2 (Pp.textf "let %s = " name ++ Dyn.pp dyn)
 ;;
 
-let rule sctx ~requires_link (exes : Dune_file.Executables.t) =
+let rule sctx ~requires_link =
   let open Action_builder.O in
   let* () = Action_builder.return () in
   let* locals, externals =
@@ -15,18 +15,17 @@ let rule sctx ~requires_link (exes : Dune_file.Executables.t) =
       | Some x -> Left x
       | None -> Right lib)
   in
+  let windows_system_values = [ "win32"; "win64"; "mingw"; "mingw64" ] in
+  let build_flags =
+    [ windows_system_values, [ "-ccopt"; "-D_UNICODE"; "-ccopt"; "-DUNICODE" ] ]
+  in
   let link_flags =
-    let win_link_flags =
-      [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ]
-    in
     (* additional link flags keyed by the platform *)
-    [ ( "macosx"
+    [ ( [ "macosx" ]
       , [ "-cclib"; "-framework CoreFoundation"; "-cclib"; "-framework CoreServices" ] )
-    ; "win32", win_link_flags
-    ; "win64", win_link_flags
-    ; "mingw", win_link_flags
-    ; "mingw64", win_link_flags
-    ; "beos", [ "-cclib"; "-lbsd" ] (* flags for Haiku *)
+    ; ( windows_system_values
+      , [ "-cclib"; "-lshell32"; "-cclib"; "-lole32"; "-cclib"; "-luuid" ] )
+    ; [ "beos" ], [ "-cclib"; "-lbsd" ] (* flags for Haiku *)
     ]
   in
   let+ locals =
@@ -66,13 +65,6 @@ let rule sctx ~requires_link (exes : Dune_file.Executables.t) =
        (Pp.concat
           ~sep:Pp.cut
           [ def
-              "executables"
-              (List
-                 (* @@DRA Want to be using the public_name here, not the
-                    internal name *)
-                 (List.map ~f:(fun (_, x) -> Dyn.String x) exes.names))
-          ; Pp.nop
-          ; def
               "external_libraries"
               (List
                  (List.filter_map externals ~f:(fun x ->
@@ -83,14 +75,19 @@ let rule sctx ~requires_link (exes : Dune_file.Executables.t) =
           ; Pp.nop
           ; def "local_libraries" (List locals)
           ; Pp.nop
+          ; def
+              "build_flags"
+              (let open Dyn in
+               list (pair (list string) (list string)) build_flags)
+          ; Pp.nop
           ; def
               "link_flags"
               (let open Dyn in
-               list (pair string (list string)) link_flags)
+               list (pair (list string) (list string)) link_flags)
           ]))
 ;;
 
-let gen_rules sctx (exes : Dune_file.Executables.t) ~dir ~requires_link =
+let gen_rules sctx (exes : Executables.t) ~dir ~requires_link =
   Memo.Option.iter exes.bootstrap_info ~f:(fun fname ->
     Super_context.add_rule
       sctx
@@ -98,5 +95,5 @@ let gen_rules sctx (exes : Dune_file.Executables.t) ~dir ~requires_link =
       ~dir
       (Action_builder.write_file_dyn
          (Path.Build.relative dir fname)
-         (rule sctx ~requires_link exes)))
+         (rule sctx ~requires_link)))
 ;;
