@@ -4,7 +4,7 @@ open Memo.O
 let ocaml_flags t ~dir (spec : Ocaml_flags.Spec.t) =
   let* expander = Super_context.expander t ~dir in
   let* flags =
-    let+ ocaml_flags = Super_context.env_node t ~dir >>= Env_node.ocaml_flags in
+    let+ ocaml_flags = Ocaml_flags_db.ocaml_flags_env ~dir in
     Ocaml_flags.make
       ~spec
       ~default:ocaml_flags
@@ -55,8 +55,8 @@ let with_lib_deps (t : Context.t) compile_info ~dir ~f =
 ;;
 
 type kind =
-  | Executables of Dune_file.Buildable.t * (Loc.t * string) list
-  | Library of Dune_file.Buildable.t * Lib_name.Local.t
+  | Executables of Buildable.t * (Loc.t * string) list
+  | Library of Buildable.t * Lib_name.Local.t
   | Melange of
       { preprocess : Preprocess.With_instrumentation.t Preprocess.Per_module.t
       ; preprocessor_deps : Dep_conf.t list
@@ -89,7 +89,7 @@ let modules_rules
       Resolve.Memo.read_memo
         (Preprocess.Per_module.instrumentation_deps preprocess ~instrumentation_backend)
     in
-    Preprocessing.make
+    Pp_spec_rules.make
       sctx
       ~dir
       ~scope
