@@ -24,6 +24,7 @@ module Expander = Expander
 module Lib = Lib
 module Lib_flags = Lib_flags
 module Lib_info = Lib_info
+module Lib_id = Lib_id
 module Modules = Modules
 module Module_compilation = Module_compilation
 module Exe_rules = Exe_rules
@@ -52,12 +53,12 @@ module Dep_rules = Dep_rules
 module Dep_graph = Dep_graph
 module Lib_config = Lib_config
 module Preprocess = Preprocess
-module Preprocessing = Preprocessing
+module Pp_spec = Pp_spec
+module Pp_spec_rules = Pp_spec_rules
 module Command = Command
 module Clflags = Clflags
 module Dune_project = Dune_project
 module Dune_project_name = Dune_project_name
-module Vcs_db = Vcs_db
 module Source_tree = Source_tree
 module Source_dir_status = Source_dir_status
 module Dune_file0 = Dune_file0
@@ -66,6 +67,8 @@ module Dialect = Dialect
 module Private_context = Private_context
 module Odoc = Odoc
 module Library = Library
+module Melange = Melange
+module Melange_stanzas = Melange_stanzas
 module Executables = Executables
 module Tests = Tests
 module Stanzas = Stanzas
