@@ -119,7 +119,7 @@ let to_dune_library (t : Findlib.Package.t) ~dir_contents ~ext_lib ~external_loc
     let implements = None in
     let orig_src_dir = None in
     let main_module_name : Lib_info.Main_module_name.t = This None in
-    let enabled = Lib_info.Enabled_status.Normal in
+    let enabled = Memo.return Lib_info.Enabled_status.Normal in
     let requires =
       Findlib.Package.requires t
       |> List.map ~f:(fun name -> Lib_dep.direct (add_loc name))
@@ -199,17 +199,20 @@ let to_dune_library (t : Findlib.Package.t) ~dir_contents ~ext_lib ~external_loc
                   then Ok None
                   else (
                     match
-                      let name = Filename.chop_extension fname in
+                      let name = Filename.remove_extension fname in
                       Module_name.of_string_user_error (Loc.in_dir src_dir, name)
                     with
                     | Ok s -> Ok (Some s)
                     | Error e -> Error e))))
     in
     let modules = Lib_info.Source.External None in
+    let name = t.name in
+    let lib_id = Lib_id.External (loc, name) in
     Lib_info.create
       ~loc
       ~path_kind:External
-      ~name:t.name
+      ~name
+      ~lib_id
       ~kind
       ~status
       ~src_dir
@@ -261,9 +264,8 @@ module Loader = struct
     in
     let rec loop ~loc ~full_name (meta : Meta.Simplified.t) acc =
       let vars = Vars.of_meta_rules meta.vars in
-      let pkg_dir = Vars.get vars "directory" Ps.empty in
       let external_location : Dune_package.External_location.t =
-        match pkg_dir with
+        match Vars.get vars "directory" Ps.empty with
         | None | Some "" -> loc
         | Some pkg_dir ->
           if pkg_dir.[0] = '+' || pkg_dir.[0] = '^'
@@ -278,15 +280,18 @@ module Loader = struct
               Relative_to_stdlib (Path.Local.relative sub pkg_dir))
           else Absolute (Path.of_filename_relative_to_initial_cwd pkg_dir)
       in
-      let dir = dir_of_loc external_location in
-      let pkg : Findlib.Package.t =
-        { Findlib.Package.meta_file; name = full_name; dir; vars }
-      in
-      let* lib =
-        let+ dir_contents = Fs.dir_contents pkg.dir in
-        to_dune_library pkg ~dir_contents ~ext_lib:db.ext_lib ~external_location
-      in
       let* (entry : Dune_package.Entry.t) =
+        let pkg : Findlib.Package.t =
+          { Findlib.Package.meta_file
+          ; name = full_name
+          ; dir = dir_of_loc external_location
+          ; vars
+          }
+        in
+        let* lib =
+          let+ dir_contents = Fs.dir_contents pkg.dir in
+          to_dune_library pkg ~dir_contents ~ext_lib:db.ext_lib ~external_location
+        in
         let+ exists =
           Findlib.Package.exists
             pkg
@@ -306,14 +311,11 @@ module Loader = struct
     in
     let name = Option.value_exn meta.name in
     let+ entries = loop ~loc ~full_name:name meta Lib_name.Map.empty in
-    let dir = dir_of_loc loc in
     { Dune_package.name = Lib_name.package_name name
     ; version =
-        (let open Option.O in
-         let* e = Lib_name.Map.find entries name in
-         Dune_package.Entry.version e)
+        Lib_name.Map.find entries name |> Option.bind ~f:Dune_package.Entry.version
     ; entries
-    ; dir
+    ; dir = dir_of_loc loc
     ; sections = Section.Map.empty
     ; sites = Site.Map.empty
     ; files = []
@@ -353,25 +355,20 @@ module Loader = struct
     | Some pkg -> Memo.return (Some (Ok pkg))
     | None ->
       let dir = Path.relative findlib_dir (Package.Name.to_string name) in
-      Fs.dir_exists dir
+      (let dune = Path.relative dir Dune_package.fn in
+       Fs.file_exists dune
+       >>= function
+       | true -> Dune_package.Or_meta.load dune
+       | false -> Memo.return (Ok Dune_package.Or_meta.Use_meta))
       >>= (function
-       | false -> Memo.return None
-       | true ->
-         (let dune = Path.relative dir Dune_package.fn in
-          Fs.file_exists dune
-          >>= function
-          | true -> Dune_package.Or_meta.load dune
-          | false -> Memo.return (Ok Dune_package.Or_meta.Use_meta))
-         >>= (function
-          | Error e ->
-            Memo.return (Some (Error (Unavailable_reason.Invalid_dune_package e)))
-          | Ok (Dune_package.Or_meta.Dune_package p) -> Memo.return (Some (Ok p))
-          | Ok Use_meta ->
-            Path.relative dir Findlib.Package.meta_fn
-            |> load_meta
-                 ~findlib_dir
-                 ~dir:(Path.Local.of_string (Package.Name.to_string name))
-            >>| Option.map ~f:(fun pkg -> Ok pkg)))
+       | Error e -> Memo.return (Some (Error (Unavailable_reason.Invalid_dune_package e)))
+       | Ok (Dune_package.Or_meta.Dune_package p) -> Memo.return (Some (Ok p))
+       | Ok Use_meta ->
+         Path.relative dir Findlib.Package.meta_fn
+         |> load_meta
+              ~findlib_dir
+              ~dir:(Path.Local.of_string (Package.Name.to_string name))
+         >>| Option.map ~f:(fun pkg -> Ok pkg))
   ;;
 
   let lookup_and_load (db : DB.t) name =
@@ -435,12 +432,12 @@ module Public = struct
   open Memo.O
 
   let find t name =
-    let+ p = find_root_package t (Lib_name.package_name name) in
-    let open Result.O in
-    let* p = p in
-    match Lib_name.Map.find p.entries name with
-    | Some x -> Ok x
-    | None -> Error Unavailable_reason.Not_found
+    Lib_name.package_name name
+    |> find_root_package t
+    >>| Result.bind ~f:(fun (p : Dune_package.t) ->
+      match Lib_name.Map.find p.entries name with
+      | Some x -> Ok x
+      | None -> Error Unavailable_reason.Not_found)
   ;;
 
   let load_all_packages (t : DB.t) =
@@ -452,23 +449,23 @@ module Public = struct
   ;;
 
   let all_packages t =
-    let+ root_packages = load_all_packages t in
-    List.fold_left root_packages ~init:[] ~f:(fun acc (_, x) ->
+    load_all_packages t
+    >>| List.fold_left ~init:[] ~f:(fun acc (_, x) ->
       match x with
       | Ok (p : Dune_package.t) ->
         Lib_name.Map.fold p.entries ~init:acc ~f:(fun x acc -> x :: acc)
       | Error _ -> acc)
-    |> List.sort ~compare:(fun a b ->
+    >>| List.sort ~compare:(fun a b ->
       Lib_name.compare (Dune_package.Entry.name a) (Dune_package.Entry.name b))
   ;;
 
   let all_broken_packages t =
-    let+ packages = load_all_packages t in
-    List.fold_left packages ~init:[] ~f:(fun acc (name, x) ->
+    load_all_packages t
+    >>| List.fold_left ~init:[] ~f:(fun acc (name, x) ->
       match x with
       | Ok _ | Error Unavailable_reason.Not_found -> acc
       | Error (Invalid_dune_package exn) -> (name, exn) :: acc)
-    |> List.sort ~compare:(fun (a, _) (b, _) -> Package.Name.compare a b)
+    >>| List.sort ~compare:(fun (a, _) (b, _) -> Package.Name.compare a b)
   ;;
 end
 
@@ -479,14 +476,17 @@ end
 type t = DB.t
 
 let create =
-  Context.DB.create_db ~name:"findlib" (fun context ->
-    let open Memo.O in
-    let* paths = Context.findlib_paths context
-    and* lib_config =
-      let+ ocaml = Context.ocaml context in
-      ocaml.lib_config
-    in
-    DB.create ~paths ~lib_config)
+  Per_context.create_by_name ~name:"findlib" (fun context ->
+    Memo.lazy_ (fun () ->
+      let open Memo.O in
+      let* context = Context.DB.get context in
+      let* paths = Context.findlib_paths context
+      and* lib_config =
+        let+ ocaml = Context.ocaml context in
+        ocaml.lib_config
+      in
+      DB.create ~paths ~lib_config)
+    |> Memo.Lazy.force)
   |> Staged.unstage
 ;;
 
