@@ -16,7 +16,7 @@ let generate_and_compile_module cctx ~precompiled_cmi ~obj_name ~name ~lib ~code
       | Some _ -> obj_name
       | None ->
         Option.map modules ~f:(fun modules ->
-          Modules.find modules name |> Option.value_exn |> Module.obj_name)
+          Modules.With_vlib.find modules name |> Option.value_exn |> Module.obj_name)
     in
     let src_dir =
       let obj_dir = Compilation_context.obj_dir cctx in
@@ -121,7 +121,7 @@ let build_info_code cctx ~libs ~api_version =
          var, placeholders)
   in
   let version_of_package placeholders (p : Package.t) =
-    match p.version with
+    match Package.version p with
     | Some v -> Memo.return (sprintf "Some %S" (Package_version.to_string v), placeholders)
     | None -> placeholder placeholders (Package.dir p)
   in
@@ -235,10 +235,7 @@ let findlib_dynload cctx lib loc =
     let open Memo.O in
     (* This shouldn't fail since findlib.dynload depends on dynlink and
        findlib. That's why it's ok to use a dummy location. *)
-    let* db =
-      let ctx = Compilation_context.context cctx in
-      Scope.DB.public_libs ctx
-    in
+    let* db = Compilation_context.context cctx |> Context.name |> Scope.DB.public_libs in
     let open Resolve.Memo.O in
     let+ dynlink = Lib.DB.resolve db (loc, Lib_name.of_string "dynlink")
     and+ findlib = Lib.DB.resolve db (loc, Lib_name.of_string "findlib") in
