@@ -14,12 +14,7 @@ let force_read_cmi source_file = [ "-intf-suffix"; Path.extension source_file ]
    generation *)
 
 let opens modules m =
-  match Modules.local_open modules m with
-  | [] -> Command.Args.empty
-  | modules ->
-    Command.Args.S
-      (List.map modules ~f:(fun name ->
-         Command.Args.As [ "-open"; Module_name.to_string name ]))
+  Command.Args.As (Modules.With_vlib.local_open modules m |> Ocaml_flags.open_flags)
 ;;
 
 let other_cm_files ~opaque ~cm_kind ~obj_dir =
@@ -60,7 +55,7 @@ let melange_args (cctx : Compilation_context.t) (cm_kind : Lib_mode.Cm_kind.t) m
       let package_output =
         Module.file ~ml_kind:Impl module_ |> Option.value_exn |> Path.parent_exn
       in
-      match Compilation_context.public_lib_name cctx with
+      match Compilation_context.melange_package_name cctx with
       | None -> [], package_output
       | Some lib_name ->
         let dir =
@@ -196,7 +191,14 @@ let build_cm
          let fn =
            Option.value_exn (Obj_dir.Module.cmt_file obj_dir m ~cm_kind ~ml_kind)
          in
-         fn :: other_targets, A "-bin-annot")
+         let annots =
+           [ "-bin-annot" ]
+           @
+           if Version.supports_bin_annot_occurrences ocaml.version
+           then [ "-bin-annot-occurrences" ]
+           else []
+         in
+         fn :: other_targets, As annots)
        else other_targets, Command.Args.empty
    in
    let opaque_arg : _ Command.Args.t =
@@ -426,7 +428,7 @@ module Alias_module = struct
     let aliases =
       Modules.Group.for_alias group
       |> List.map ~f:(fun (local_name, m) ->
-        let canonical_path = Modules.canonical_path modules group m in
+        let canonical_path = Modules.With_vlib.canonical_path modules group m in
         let obj_name = Module.obj_name m in
         { canonical_path; local_name; obj_name })
     in
@@ -504,7 +506,7 @@ let build_all cctx =
   let for_wrapped_compat = lazy (Compilation_context.for_wrapped_compat cctx) in
   let modules = Compilation_context.modules cctx in
   Memo.parallel_iter
-    (Modules.fold_no_vlib_with_aliases
+    (Modules.With_vlib.fold_no_vlib_with_aliases
        modules
        ~init:[]
        ~normal:(fun x acc -> `Normal x :: acc)
@@ -520,7 +522,7 @@ let build_all cctx =
            build_module cctx m
          | _ ->
            let cctx =
-             if Modules.is_stdlib_alias modules m
+             if Modules.With_vlib.is_stdlib_alias modules m
              then
                (* XXX it would probably be simpler if the flags were just for this
                   module in the definition of the stanza *)
