@@ -144,9 +144,73 @@ module Install_cookie = struct
   ;;
 end
 
+module Value_list_env = struct
+  (* A representation of an environment where each variable can hold a
+     list of [Value.t]. Each variable will be encoded into a delimited
+     string (in the style of the PATH variable). *)
+  type t = Value.t list Env.Map.t
+
+  let parse_strings s = Bin.parse s |> List.map ~f:(fun s -> Value.String s)
+  let of_env env : t = Env.to_map env |> Env.Map.map ~f:parse_strings
+
+  (* Concatenate a list of values in the style of lists found in
+     environment variables, such as PATH *)
+  let string_of_env_values values =
+    List.map values ~f:(function
+      | Value.String s -> s
+      | Dir s | Path s -> Path.to_absolute_filename s)
+    |> Bin.encode_strings
+  ;;
+
+  let to_env (t : t) = Env.Map.map t ~f:string_of_env_values |> Env.of_map
+  let get_path t = Env.Map.find t Env_path.var
+
+  (* [extend_concat_path a b] adds all variables from [b] to [a]
+     overwriting any existing values of those variables in [a] except for PATH
+     which is set to the concatenation of the PATH variables from [a] and [b]
+     with the PATH entries from [b] preceding the PATH entries from
+     [a]. If only one of the arguments contains a PATH variable then
+     its value will be the value of PATH in the result, however if
+     neither argument contains a PATH variable then PATH will be unset
+     in the result. *)
+  let extend_concat_path a b =
+    let extended = Env.Map.superpose b a in
+    let concated_path =
+      match get_path a, get_path b with
+      | None, None -> None
+      | Some x, None | None, Some x -> Some x
+      | Some a, Some b -> Some (b @ a)
+    in
+    match concated_path with
+    | None -> extended
+    | Some concated_path -> Env.Map.set extended Env_path.var concated_path
+  ;;
+
+  (* Adds a path to an env where variables are associated with lists
+     of paths. The path is prepended to the list associated with the
+     given variable and a new binding is added to the env if the
+     variable is not yet part of the env. *)
+  let add_path (t : t) var path : t =
+    Env.Map.update t var ~f:(fun paths ->
+      let paths = Option.value paths ~default:[] in
+      Some (Value.Dir (Path.build path) :: paths))
+  ;;
+end
+
 module Env_update = struct
   include Dune_lang.Action.Env_update
 
+  (* Handle the :=, +=, =:, and =+ opam environment update operators.
+
+     The operators with colon character update a variable, adding a
+     leading/trailing separator (e.g. the ':' chars in PATH on unix)
+     if the variable was initially unset or empty, while the operators
+     with a plus character add no leading/trailing separator in such a
+     case.
+
+     Updates where the newly added value is the empty string are
+     ignored since opam refuses to add empty strings to list
+     variables.*)
   let update kind ~new_v ~old_v ~f =
     if new_v = ""
     then old_v
@@ -181,13 +245,6 @@ module Env_update = struct
         (* TODO nobody uses this AFAIK *)
         assert false)
   ;;
-
-  let string_of_env_values values =
-    List.map values ~f:(function
-      | Value.String s -> s
-      | Dir s | Path s -> Path.to_absolute_filename s)
-    |> Bin.encode_strings
-  ;;
 end
 
 module Pkg = struct
@@ -257,10 +314,16 @@ module Pkg = struct
         in
         Path.Local.Set.union_all (acc :: dirs)
     in
-    match t.info.source with
+    (match t.info.source with
+     | None -> Memo.return None
+     | Some source ->
+       Lock_dir.source_kind source
+       >>| (function
+        | `Local (`File, _) | `Fetch -> None
+        | `Local (`Directory, root) -> Some root))
+    >>= function
     | None -> Memo.return Path.Local.Set.empty
-    | Some (External_copy (_, root)) -> loop root Path.Local.Set.empty Path.Local.root
-    | Some (Fetch _) -> assert false
+    | Some root -> loop root Path.Local.Set.empty Path.Local.root
   ;;
 
   let dep t = Dep.file (Path.build t.paths.target_dir)
@@ -270,44 +333,47 @@ module Pkg = struct
     |> List.fold_left ~init:Dep.Set.empty ~f:(fun acc t -> dep t |> Dep.Set.add acc)
   ;;
 
-  let build_env_of_deps =
-    let add_to_path env var what =
-      Env.Map.update env var ~f:(fun paths ->
-        let paths = Option.value paths ~default:[] in
-        Some (Value.Dir (Path.build what) :: paths))
-    in
-    fun xs ->
-      List.fold_left xs ~init:Env.Map.empty ~f:(fun env t ->
-        let env =
-          let roots = Paths.install_roots t.paths in
-          let init = add_to_path env Env_path.var roots.bin in
-          let vars = Install.Roots.to_env_without_path roots in
-          List.fold_left vars ~init ~f:(fun acc (var, path) -> add_to_path acc var path)
-        in
-        List.fold_left t.exported_env ~init:env ~f:Env_update.set)
+  (* Given a list of packages, construct an env containing variables
+     set by each package. Variables containing delimited lists of
+     paths (e.g. PATH) which appear in multiple package's envs are
+     concatenated in the reverse order of their associated packages in
+     the input list. Environment updates via the `exported_env` field
+     (equivalent to opam's `setenv` field) are applied for each
+     package in the same order as the argument list. *)
+  let build_env_of_deps ts =
+    List.fold_left ts ~init:Env.Map.empty ~f:(fun env t ->
+      let env =
+        let roots = Paths.install_roots t.paths in
+        let init = Value_list_env.add_path env Env_path.var roots.bin in
+        let vars = Install.Roots.to_env_without_path roots in
+        List.fold_left vars ~init ~f:(fun acc (var, path) ->
+          Value_list_env.add_path acc var path)
+      in
+      List.fold_left t.exported_env ~init:env ~f:Env_update.set)
   ;;
 
+  (* [build_env t] returns an env containing paths containing all the
+     tools and libraries required to build the package [t] inside the
+     faux opam directory contained in the _build dir. *)
   let build_env t = build_env_of_deps @@ deps_closure t
 
-  let exported_env t =
-    let base =
-      Env.Map.of_list_exn
-        [ Opam_switch.opam_switch_prefix_var_name, Path.Build.to_string t.paths.target_dir
-        ; "CDPATH", ""
-        ; "MAKELEVEL", ""
-        ; "OPAM_PACKAGE_NAME", Package.Name.to_string t.info.name
-        ; "OPAM_PACKAGE_VERSION", Package_version.to_string t.info.version
-        ; "OPAMCLI", "2.0"
-        ]
-    in
-    let package_env =
-      let vars =
-        build_env t
-        |> Env.Map.map ~f:Env_update.string_of_env_values
-        |> Env.Map.superpose base
-      in
-      Env.extend Env.empty ~vars
-    in
+  let base_env t =
+    Env.Map.of_list_exn
+      [ ( Opam_switch.opam_switch_prefix_var_name
+        , [ Value.Path (Path.build t.paths.target_dir) ] )
+      ; "CDPATH", [ Value.String "" ]
+      ; "MAKELEVEL", [ Value.String "" ]
+      ; "OPAM_PACKAGE_NAME", [ Value.String (Package.Name.to_string t.info.name) ]
+      ; ( "OPAM_PACKAGE_VERSION"
+        , [ Value.String (Package_version.to_string t.info.version) ] )
+      ; "OPAMCLI", [ Value.String "2.0" ]
+      ]
+  ;;
+
+  (* [exported_value_env t] returns the complete env that will be used
+     to build the package [t] *)
+  let exported_value_env t =
+    let package_env = build_env t |> Env.Map.superpose (base_env t) in
     (* TODO: Run actions in a constrained environment. [Global.env ()] is the
        environment from which dune was executed, and some of the environment
        variables may affect builds in unintended ways and make builds less
@@ -315,8 +381,10 @@ module Pkg = struct
        for build actions to run successfully, such as $PATH on systems where the
        shell's default $PATH variable doesn't include the location of standard
        programs or build tools (e.g. NixOS). *)
-    Env_path.extend_env_concat_path (Global.env ()) package_env
+    Value_list_env.extend_concat_path (Value_list_env.of_env (Global.env ())) package_env
   ;;
+
+  let exported_env t = Value_list_env.to_env @@ exported_value_env t
 end
 
 module Pkg_installed = struct
@@ -456,6 +524,7 @@ module Run_with_path = struct
     type ('path, 'target) t =
       { prog : Action.Prog.t
       ; args : 'path arg Array.Immutable.t
+      ; ocamlfind_destdir : 'path
       }
 
     let name = "run-with-path"
@@ -467,10 +536,16 @@ module Run_with_path = struct
         | Path p -> Path (f p))
     ;;
 
-    let bimap t f _g = { t with args = Array.Immutable.map t.args ~f:(map_arg ~f) }
+    let bimap t f _g =
+      { t with
+        args = Array.Immutable.map t.args ~f:(map_arg ~f)
+      ; ocamlfind_destdir = f t.ocamlfind_destdir
+      }
+    ;;
+
     let is_useful_to ~memoize:_ = true
 
-    let encode { prog; args } path _ : Dune_lang.t =
+    let encode { prog; args; ocamlfind_destdir } path _ : Dune_lang.t =
       let prog =
         Dune_lang.atom_or_quoted_string
         @@
@@ -485,10 +560,17 @@ module Run_with_path = struct
               | String s -> Dune_lang.atom_or_quoted_string s
               | Path p -> path p)))
       in
-      List ([ Dune_lang.atom_or_quoted_string name; prog ] @ args)
+      List
+        [ List ([ Dune_lang.atom_or_quoted_string name; prog ] @ args)
+        ; path ocamlfind_destdir
+        ]
     ;;
 
-    let action { prog; args } ~(ectx : Action.Ext.context) ~(eenv : Action.Ext.env) =
+    let action
+      { prog; args; ocamlfind_destdir }
+      ~(ectx : Action.Ext.context)
+      ~(eenv : Action.Ext.env)
+      =
       let open Fiber.O in
       match prog with
       | Error e -> Action.Prog.Not_found.raise e
@@ -501,6 +583,12 @@ module Run_with_path = struct
             |> String.concat ~sep:"")
         in
         let metadata = Process.create_metadata ~purpose:ectx.purpose () in
+        let env =
+          Env.add
+            eenv.env
+            ~var:"OCAMLFIND_DESTDIR"
+            ~value:(Path.to_absolute_filename ocamlfind_destdir)
+        in
         Process.run
           (Accept eenv.exit_codes)
           prog
@@ -511,21 +599,21 @@ module Run_with_path = struct
           ~stderr_to:eenv.stderr_to
           ~stdin_from:eenv.stdin_from
           ~dir:eenv.working_dir
-          ~env:eenv.env
+          ~env
         >>= (function
          | Error _ -> Fiber.return ()
          | Ok () -> Fiber.return ())
     ;;
   end
 
-  let action prog args =
+  let action prog args ~ocamlfind_destdir =
     let module M = struct
       type path = Path.t
       type target = Path.Build.t
 
       module Spec = Spec
 
-      let v = { Spec.prog; args }
+      let v = { Spec.prog; args; ocamlfind_destdir }
     end
     in
     Action.Extension (module M)
@@ -786,7 +874,10 @@ module Action_expander = struct
                | String s -> Run_with_path.Spec.String s
                | Path p | Dir p -> Path p))
          in
-         Run_with_path.action exe args)
+         let ocamlfind_destdir =
+           (Lazy.force expander.paths.install_roots).lib_root |> Path.build
+         in
+         Run_with_path.action exe args ~ocamlfind_destdir)
     | Progn t ->
       let+ args = Memo.parallel_map t ~f:(expand ~expander) in
       Action.Progn args
@@ -850,7 +941,7 @@ module Action_expander = struct
           let update =
             let value =
               match Env.Map.find env var with
-              | Some v -> Env_update.string_of_env_values v
+              | Some v -> Value_list_env.string_of_env_values v
               | None ->
                 (* TODO *)
                 ""
@@ -908,7 +999,7 @@ module Action_expander = struct
     let+ { Artifacts_and_deps.binaries; dep_info } =
       Pkg.deps_closure pkg |> Artifacts_and_deps.of_closure
     in
-    let env = Pkg.build_env pkg in
+    let env = Pkg.exported_value_env pkg in
     let depends =
       Package.Name.Map.add_exn
         dep_info
@@ -958,7 +1049,7 @@ module Action_expander = struct
   ;;
 
   let install_command context (pkg : Pkg.t) =
-    Option.map pkg.install_command ~f:(expand context pkg)
+    Option.map pkg.install_command ~f:(fun action -> expand context pkg action)
   ;;
 
   let exported_env (expander : Expander.t) (env : _ Env_update.t) =
@@ -1014,10 +1105,7 @@ end = struct
         let+ lock_dir = Lock_dir.get_path ctx >>| Option.value_exn in
         Path.Build.append_source
           (Context_name.build_dir ctx)
-          (Path.Source.relative
-             lock_dir
-             (* TODO this should come from [Dune_pkg] *)
-             (sprintf "%s.files" (Package.Name.to_string info.name)))
+          (Dune_pkg.Lock_dir.Pkg.files_dir info.name ~lock_dir)
       in
       let id = Pkg.Id.gen () in
       let paths = Paths.make name ctx in
@@ -1227,9 +1315,12 @@ module Install_action = struct
       | false -> []
       | true ->
         let config =
-          let config_file_str = Path.to_string config_file in
-          let file = OpamFilename.of_string config_file_str |> OpamFile.make in
-          match OpamFile.Dot_config.read file with
+          let filename = Path.to_string config_file in
+          match
+            Io.read_file config_file
+            |> OpamFile.Dot_config.read_from_string
+                 ~filename:(OpamFile.make (OpamFilename.of_string filename))
+          with
           | s -> s
           | exception OpamPp.Bad_format (pos, message) ->
             let loc =
@@ -1243,7 +1334,7 @@ module Install_action = struct
                   let bols = Array.of_list (List.rev !bols) in
                   let make_pos (line, column) =
                     let pos_bol = bols.(line - 1) in
-                    { Lexing.pos_fname = config_file_str
+                    { Lexing.pos_fname = filename
                     ; pos_lnum = line
                     ; pos_bol
                     ; pos_cnum = pos_bol + column
@@ -1400,32 +1491,37 @@ let source_rules (pkg : Pkg.t) =
   let+ source_deps, copy_rules =
     match pkg.info.source with
     | None -> Memo.return (Dep.Set.empty, [])
-    | Some (Fetch { url = (loc, _) as url; checksum }) ->
-      let fetch =
-        Fetch_rules.fetch ~target:pkg.paths.source_dir `Directory url checksum
-      in
-      Memo.return (Dep.Set.of_files [ Path.build pkg.paths.source_dir ], [ loc, fetch ])
-    | Some (External_copy (loc, source_root)) ->
-      let+ source_files, rules =
-        let source_root = Path.external_ source_root in
-        Pkg.source_files pkg ~loc
-        >>| Path.Local.Set.fold ~init:([], []) ~f:(fun file (source_files, rules) ->
-          let src = Path.append_local source_root file in
-          let dst = Path.Build.append_local pkg.paths.source_dir file in
-          let copy = loc, Action_builder.copy ~src ~dst in
-          Path.build dst :: source_files, copy :: rules)
-      in
-      Dep.Set.of_files source_files, rules
+    | Some source ->
+      let loc = fst source.url in
+      Lock_dir.source_kind source
+      >>= (function
+       | `Local (`File, _) | `Fetch ->
+         let fetch = Fetch_rules.fetch ~target:pkg.paths.source_dir `Directory source in
+         Memo.return (Dep.Set.of_files [ Path.build pkg.paths.source_dir ], [ loc, fetch ])
+       | `Local (`Directory, source_root) ->
+         let+ source_files, rules =
+           let source_root = Path.external_ source_root in
+           Pkg.source_files pkg ~loc
+           >>| Path.Local.Set.fold ~init:([], []) ~f:(fun file (source_files, rules) ->
+             let src = Path.append_local source_root file in
+             let dst = Path.Build.append_local pkg.paths.source_dir file in
+             let copy = loc, Action_builder.copy ~src ~dst in
+             Path.build dst :: source_files, copy :: rules)
+         in
+         Dep.Set.of_files source_files, rules)
   in
   let extra_source_deps, extra_copy_rules =
-    List.map pkg.info.extra_sources ~f:(fun (local, fetch) ->
+    List.map pkg.info.extra_sources ~f:(fun (local, (fetch : Source.t)) ->
       let extra_source = Paths.extra_source pkg.paths local in
       let rule =
-        match (fetch : Source.t) with
-        | External_copy (loc, src) ->
+        let loc = fst fetch.url in
+        (* We assume that [fetch] is always a file. Would be good
+           to give a decent error message if it's not *)
+        match Source.kind fetch with
+        | `Directory_or_archive src ->
           loc, Action_builder.copy ~src:(Path.external_ src) ~dst:extra_source
-        | Fetch { url = (loc, _) as url; checksum } ->
-          let rule = Fetch_rules.fetch ~target:pkg.paths.source_dir `File url checksum in
+        | `Fetch ->
+          let rule = Fetch_rules.fetch ~target:pkg.paths.source_dir `File fetch in
           loc, rule
       in
       Path.build extra_source, rule)
@@ -1526,8 +1622,8 @@ module Gen_rules = Build_config.Gen_rules
 
 let setup_package_rules context ~dir ~pkg_name : Gen_rules.result Memo.t =
   let name = User_error.ok_exn (Package.Name.of_string_user_error (Loc.none, pkg_name)) in
-  let* db = DB.get context in
-  let+ pkg =
+  let* pkg =
+    let* db = DB.get context in
     Resolve.resolve db context (Loc.none, name)
     >>| function
     | `Inside_lock_dir pkg -> pkg
@@ -1540,12 +1636,19 @@ let setup_package_rules context ~dir ~pkg_name : Gen_rules.result Memo.t =
         ]
   in
   let paths = Paths.make name context in
-  let directory_targets =
-    let target_dir = paths.target_dir in
-    let map = Path.Build.Map.singleton target_dir Loc.none in
+  let+ directory_targets =
+    let map =
+      let target_dir = paths.target_dir in
+      Path.Build.Map.singleton target_dir Loc.none
+    in
     match pkg.info.source with
-    | Some (Fetch f) -> Path.Build.Map.add_exn map paths.source_dir (fst f.url)
-    | _ -> map
+    | None -> Memo.return map
+    | Some source ->
+      Lock_dir.source_kind source
+      >>| (function
+       | `Local (`Directory, _) -> map
+       | `Local (`File, _) | `Fetch ->
+         Path.Build.Map.add_exn map paths.source_dir (fst source.url))
   in
   let build_dir_only_sub_dirs =
     Gen_rules.Build_only_sub_dirs.singleton ~dir Subdir_set.empty
@@ -1638,7 +1741,7 @@ let lock_dir_path = Lock_dir.get_path
 let exported_env context =
   let+ all_packages = all_packages context in
   let env = Pkg.build_env_of_deps all_packages in
-  let vars = Env.Map.map env ~f:Env_update.string_of_env_values in
+  let vars = Env.Map.map env ~f:Value_list_env.string_of_env_values in
   Env.extend Env.empty ~vars
 ;;
 
